"use strict";
this.name = "ExternalDockSystem";
this.author = "phkb";
this.description = "System for add external docks to stations";
this.version = "1.1";
this.copyright = "2024 phkb";
this.license = "CC BY-NC-SA 4.0";

this._fcb = null;
this._externalDocks = [];

//----------------------------------------------------------------------------------------
this.shipWillEnterEnterWitchspace = function() {
    this.$stopFCB();
    this._externalDocks.length = 0;
}

//----------------------------------------------------------------------------------------
this.shipWillDockWithStation = this.shipDied = function() {
    this.$stopFCB();
}

//----------------------------------------------------------------------------------------
this.shipExitedWitchspace = function() {
    if (this._externalDocks.length > 0) {
        for (var i = 0; i < this._externalDocks.length; i++) {
            this._externalDocks[i].externalDock._speedWarning = false;
        }
        this.$startFCB();
    }
}

//----------------------------------------------------------------------------------------
this.shipWillLaunchFromStation = function(station) {
    var p = player.ship;
    if (station.hasOwnProperty("_dockedViaED")) {
        var dve = station._dockedViaED;
        dve._speedWarning = true;
        p.position = dve.position.add(dve.vectorForward.multiply(p.boundingBox.z + 50));
        p.orientation = dve.orientation;
        p.velocity = p.vectorForward.multiply(p.speed);
        delete station._dockedViaED;
    }
    if (this._externalDocks.length > 0) this.$startFCB();
}

//----------------------------------------------------------------------------------------
this.$addExternalDock = function(stn, position, scale, predock_callback, postdock_callback) {
    if (!stn.isStation) throw "Invalid settings: target entity is not a station.";
    if (!Array.isArray(position)) throw "Invalid settings: position property is not an array";
    if (position.length != 3) throw "Invalid settings: position property has incorrect number of elements";
    if (scale && isNaN(scale)) throw "Invalid settings: scale property is no a valid number";

    var dve = system.addVisualEffect("external_dock_ve_blank", 
        stn.position.add(stn.vectorRight.multiply(position[0])).add(stn.vectorUp.multiply(position[1])).add(stn.vectorForward.multiply(position[2])));
    if (!scale || scale == 0) scale = 1;
    if (scale && scale != 1) dve.scale(scale);

    stn._savedPosition = stn.position;
    stn._savedOrientation = stn.orientation;
    stn._isMoving = false; // we'll change this to true if the position or orientation changes.

    this._externalDocks.push({
        station:stn, 
        externalDock:dve, 
        relativePos:position, 
        scale:scale,
        preDockCallback:predock_callback, 
        postDockCallback:postdock_callback
    });
}

//----------------------------------------------------------------------------------------
this.$startFCB = function() {
    if (!this._fcb || !isValidFrameCallback(this._fcb)) {
        this._fcb = addFrameCallback(this.$checkExternalDocks);
    }
}

//----------------------------------------------------------------------------------------
this.$stopFCB = function() {
    if (this._fcb && isValidFrameCallback(this._fcb)) removeFrameCallback(this._fcb);
    this._fcb = null;
}

//----------------------------------------------------------------------------------------
this.$checkExternalDocks = function _checkExternalDocks(delta) {
    // inline function to reposition docks on stations if they move
    function reposition(obj) {
        var dve = obj.externalDock;
        var stn = obj.station;
        var pos = obj.relativePos;

        dve.position = stn.position.add(stn.vectorRight.multiply(pos[0])).add(stn.vectorUp.multiply(pos[1])).add(stn.vectorForward.multiply(pos[2]));
        
        stn._savedPosition = stn.position;
        stn._savedOrientation = stn.orientation;
    }

    var that = _checkExternalDocks;
    var _p = (that._p = that._p || player.ship);
    var _ed = (that._ed = that._ed || worldScripts.ExternalDockSystem._externalDocks);
    var i = _ed.length;
    while (i--) {
        var ed = _ed[i].externalDock;
        var stn = _ed[i].station;
        var scale = _ed[i].scale;

        // check to see if station has moved
        // if so, perform function to adjust location of dock flashers
        if (stn._isMoving || (stn.position.angleTo(stn._savedPosition) != 0 || stn.orientation.dot(stn._savedOrientation) != 1)) {
            stn._isMoving = true;
            reposition(_ed[i]);
        }

        // check player distance to dock
        var dist = ed.position.distanceTo(_p);
        if (dist > (2800 * scale) && ed._speedWarning) ed._speedWarning = false;
        if (dist <= (2000 * scale) && _p.speed < 100 && ed._speedWarning) ed._speedWarning = false;
        if (dist <= (2000 * scale) && _p.speed >= 100 && !ed._speedWarning) {
            ed._speedWarning = true;
            stn.commsMessage("Reduce speed to less than 100m/s for docking.", _p);
        }
        if (dist < 150 && _p.speed < 100) {
            stn._dockedViaED = ed;
            if (_ed[i].preDockCallback) {
                try {
                    _ed[i].preDockCallback(stn);
                } catch (ex) {
                    log("external docking", "ERROR calling pre-dock callback: " + ex);
                }
            }
            // check if the player has requested docking clearance (and the station requires it)
            // if they haven't requested it, fine the player
            if (stn.requiresDockingClearance && (player.dockingClearanceStatus != "DOCKING_CLEARANCE_STATUS_GRANTED" && player.dockingClearanceStatus != "DOCKING_CLEARANCE_STATUS_REQUESTED" && player.dockingClearanceStatus != "DOCKING_CLEARANCE_STATUS_TIMING_OUT")) {
                var fine = parseInt(player.credits * 0.05);
                if (fine > 5000) fine = 5000;
                player.credits -= fine;
                var txt = expandDescription("[station-docking-clearance-fined-@-cr]").replace("%@", formatCredits(fine, false, true));
                player.addMessageToArrivalReport(txt);
            }   
            stn.dockPlayer();
            if (_ed[i].postDockCallback) {
                try {
                    _ed[i].postDockCallback(stn);
                } catch (ex) {
                    log("external docking", "ERROR calling post-dock callback: " + ex);
                }
            }
            return;
        }
    }
}